/*
 * main.c
 *
 *  Created on: 16 Feb 2018
 *      Author: ctorr
 */

#pragma attribute("aid", "F5000001")
#pragma attribute("access_list", "1000")
#pragma attribute("name", "ESP01 Wifi Driver v1")

#include <multos.h>
#include <mi.h>
#include <string.h>
#include <ISO7816.h>
#include <stdlib.h>
#include <stdio.h>
#include "wifi.h"

/* This version supports the following setup of the ESP8266
 *
 * See https://www.espressif.com/sites/default/files/documentation/4a-esp8266_at_instruction_set_en.pdf
 *
AT version:1.6.0.0(Feb  3 2018 12:00:06)
SDK version:2.2.1(f42c330)
compile time:Feb 12 2018 16:31:26
Bin version(Wroom 02):1.6.1

The default baud rate needs to be changed to 57600 using the command

AT+UART_DEF=57600,8,1,0,1

The default mode needs to be set
AT+CWMODE_DEF=1
*/

// **************** MEMORY SEGMENTS **********************
#pragma melpublic
union
{
	wifiConnectApdu_t connectData;
	wifiHttpGetReqApdu_t requestData;
	mqttConnectApdu_t mqttConnectData;
	mqttPublishApdu_t mqttPublishData;
	char acReply[WIFI_MAX_REPLY];
} apduData;

#pragma melstatic
BYTE crlf[] = { 0x0D, 0x0A };

// CMD (1) | LEN (1) | PROTOCOL NAME (6) | Protocol Level (1) | Connect flags (1) | Keep Alive time (2) | Client Id Len (2) | Client Id
// Assumption is that clientId is short so that LEN remains one byte (<=127)
BYTE mqttConnectHdr[] = { 0x10, 0x00, 0x00, 0x04, 'M', 'Q', 'T', 'T', 0x04, 0x02, 0x00, 0x3C, 0x00, 0x00 };
BYTE mqttConnAckMsg[] = { 0x20, 0x02, 0x00, 0x00 };
BYTE mqttPublishHdr[] = { 0x30, 0x00, 0x00 };
BYTE mqttDisconnectMsg[] = { 0xE0, 0x00 };

#pragma melsession
BYTE bSerialPortNumber;

// ************* LOCAL FUNCTIONS ************************
void abendFnc(void)
{
	// Get abend reason code
	SW1 = 0x6A;
	SW2 = miAbendReason();
	multosExit();
}

char *findStr(char *sFull, char *sPart, WORD wPartLen)
{
	char *s;
	char *l;

	for(s = sFull; *s != NULL && l == NULL; s++)
		if(*s == sPart[0])
			if(strncmp(s,sPart,wPartLen) == 0)
				l = s;

	return(l);
}

// Returns the number of bytes in the reply
WORD sendATCmd(char *abCmd, char *abResp, WORD wMaxReplyLen, WORD wMaxWait, BOOL bIsLongCommand)
{
	WORD wCount;
	WORD w;
	WORD wWaits;

	// Transmit
	miSerialIOTransmitBytes((BYTE*)abCmd,strlen(abCmd));
	miSerialIOTransmitBytes(crlf,2);

	// Initial wait. This allows for progress messages that come back very quickly to not trigger the
	// following process too soon.
	miDelay(200);

	// Wait for reply to start
	wWaits = wMaxWait / 5;
	while(wWaits-- && miSerialIOGetNumReceivedBytes() == 0)
		miDelay(5);

	// If a reply was received
	if(wWaits)
	{
		// Read the reply up to the size of the supplied buffer leaving room for NULL termination
		while(miSerialIOGetNumReceivedBytes() && wCount < wMaxReplyLen-1)
		{
			wCount += miSerialIOGetReceivedBytes((BYTE*)(abResp + wCount),wMaxReplyLen - wCount - 1);

			// NULL terminate reply
			abResp[wCount] = 0;

			// Some commands output info messages as they go along. e.g. Wifi connection. This takes some time and
			// there may be gaps.
			if(bIsLongCommand && findStr(abResp,"OK",2) == NULL)
			{
				wWaits = wMaxWait / 50;
				while(wWaits-- && miSerialIOGetNumReceivedBytes() == 0)
					miDelay(50);
			}
		}
	}

	// Trim \r\n from the beginning and end if there
	if(wCount >= 2)
	{
		if(abResp[wCount-2] == 0x0D && abResp[wCount-1] == 0x0A)
		{
			// NULL terminate 2 characters sooner and reduce length by 2
			abResp[wCount-2] = 0;
			wCount = wCount - 2;
		}

		if(abResp[0] == 0x0D && abResp[1] == 0x0A)
		{
			// Shuffle down two bytes
			for(w = 0; w < wCount-2; w++)
				abResp[w] = abResp[w+2];

			// Terminate
			abResp[wCount-2] = 0;
			wCount = wCount - 2;
		}
	}
	return(wCount);
}

WORD readReply(char *abResp, WORD wMaxReplyLen, WORD wMaxWait, char* acCheckStr)
{
	WORD wCount;
	WORD wWaits;
	WORD wCheckStrLen;

	if(acCheckStr)
		wCheckStrLen = strlen(acCheckStr);

	// Wait for reply to start
	miDelay(200);
	wWaits = (wMaxWait-200) / 200;
	while(wWaits-- && miSerialIOGetNumReceivedBytes() == 0)
		miDelay(200);

	// If a reply was received
	if(wWaits)
	{
		// Read the reply up to the size of the supplied buffer leaving room for NULL termination
		while(miSerialIOGetNumReceivedBytes() && wCount < wMaxReplyLen-1)
		{
			wCount += miSerialIOGetReceivedBytes((BYTE*)(abResp + wCount),wMaxReplyLen - wCount - 1);

			abResp[wCount] = 0;

			// There may be gaps in transmission
			if(acCheckStr && findStr(abResp,acCheckStr,wCheckStrLen) == 0)
			{
				wWaits = wMaxWait / 50;
				while(wWaits-- && miSerialIOGetNumReceivedBytes() == 0)
					miDelay(50);
			}
		}
	}


	return(wCount);
}

// ***************** MAIN ******************************
void
main(void)
{
	WORD wReplyLen;
	WORD wReqLen;
	WORD wTotalLen;
	WORD wTopicLen;
	BYTE bLen;
	char acATCmd[261];
	char acATReply[256];


	// Register abend handler
	miSetAbendFunction((WORD)abendFnc);

	// Ignore startup event
	bLen = miGetEmbeddedEventId();
	if( bLen == MI_EMBEDDED_EVENT_STARTUP )
		return;

	// Check CLA byte OK
	if(CLA != WIFI_CLA)
		multosExitSW(ISO7816_SW_CLA_NOT_SUPPORTED);

	// Activate the correct serial port
	miSerialIOSetActivePort(bSerialPortNumber);
	miSerialIOFlushReceiveBuffer();

	// Process instructions
	switch(INS)
	{
		case WIFI_SETUP_INS:
			if(!multosCheckCase(1))
				multosExitSW(ISO7816_SW_WRONG_PARAMS);

			// Check interface type is serial and P2 is zero or 1
			if(P1 != WIFI_SERIAL_IF || P2 > 1)
				multosExitSW(ISO7816_SW_INCORRECT_P1P2);

			// Set serial port number
			bSerialPortNumber = P2;

			// Turn off AT echo
			miSerialIOSetActivePort(bSerialPortNumber);
			wReplyLen = sendATCmd("ATE0",acATReply,sizeof(acATReply),500,FALSE);
			miSerialIOFlushReceiveBuffer();
			apduData.acReply[0] = 0;
			break;

		case WIFI_CONNECT_INS:
			if(!multosCheckCase(3))
				multosExitSW(ISO7816_SW_WRONG_PARAMS);

			// Check Wifi module present and working OK
			wReplyLen = sendATCmd("AT",acATReply,sizeof(acATReply),500,FALSE);
			if(acATReply[wReplyLen-2] == 'O' && acATReply[wReplyLen-1] == 'K')
			{
				// Build AT command for connection
				acATCmd[0] = 0;
				strcat(acATCmd,"AT+CWJAP_CUR=\"");
				strcat(acATCmd,apduData.connectData.acSSID);
				strcat(acATCmd,"\",\"");
				strcat(acATCmd,apduData.connectData.acPWD);
				strcat(acATCmd,"\"");

				// Send AT command and check reply is OK
				wReplyLen = sendATCmd(acATCmd,acATReply,sizeof(acATReply),30000,TRUE);
				if(acATReply[wReplyLen-2] != 'O' || acATReply[wReplyLen-1] != 'K')
				{
					apduData.acReply[0] = 0;
					multosExitSW(WIFI_ERR_CONNECT);
				}
			}
			else
			{
				//apduData.acReply[0] = 0;
				strcpy(apduData.acReply,acATReply);
				La = wReplyLen;
				multosExitSW(WIFI_ERR_DEVICE);
			}
			apduData.acReply[0] = 0;
			break;

		case WIFI_DISCONNECT_INS:
			break;

		case WIFI_HTTPGET_INS:
			if(!multosCheckCase(4))
				multosExitSW(ISO7816_SW_WRONG_PARAMS);

			miSerialIOFlushReceiveBuffer();

			// Build AT START command
			La = 0;
			acATCmd[0] = 0;
			strcat(acATCmd,"AT+CIPSTART=\"");
			strcat(acATCmd,apduData.requestData.acProtocol);
			strcat(acATCmd,"\",\"");
			strcat(acATCmd,apduData.requestData.acServer);
			strcat(acATCmd,"\",");
			ultoa(apduData.requestData.wPort,acATReply,10);
			strcat(acATCmd,acATReply);

			// Send AT command and check reply is OK
			wReplyLen = sendATCmd(acATCmd,acATReply,sizeof(acATReply),500,FALSE);
			if(findStr(acATReply,"OK",2) == NULL)
			{
				apduData.acReply[0] = 0;
				multosExitSW(WIFI_ERR_OPEN);
			}

			// Build AT SEND command
			acATCmd[0] = 0;
			strcat(acATCmd,"AT+CIPSEND=");
			// GET <uri> HTTP/1.1\r\nHost: <server>\r\nConnection: close\r\n\r\n
			wReqLen = 4 + strlen(apduData.requestData.acUri) +  17 + strlen(apduData.requestData.acServer) + 23;
			ultoa(wReqLen,acATReply,10);
			strcat(acATCmd,acATReply);

			// Send AT command and check reply is a '>' character
			wReplyLen = sendATCmd(acATCmd,acATReply,sizeof(acATReply),500,FALSE);
			if(wReplyLen < 2 || acATReply[wReplyLen-2] != '>')
			{
				apduData.acReply[0] = 0;
				multosExitSW(WIFI_ERR_SEND);
			}

			// Transmit the HTTP GET request bytes
			miSerialIOTransmitBytes((BYTE*)"GET ",4);
			miSerialIOTransmitBytes((BYTE*)apduData.requestData.acUri,strlen(apduData.requestData.acUri));
			miSerialIOTransmitBytes((BYTE *)" HTTP/1.1\r\nHost: ",17);
			miSerialIOTransmitBytes((BYTE *)apduData.requestData.acServer,strlen(apduData.requestData.acServer));
			miSerialIOTransmitBytes((BYTE *) "\r\nConnection: close\r\n\r\n", 23);

			apduData.acReply[0] = NULL;
			wReplyLen = readReply(apduData.acReply,WIFI_MAX_REPLY-1,apduData.requestData.wMaxWait, "CLOSED");
			La = wReplyLen;

			// Should get SEND OK as part of the returned data
			if(findStr(apduData.acReply,"SEND OK",7) == NULL)
				multosExitSW(WIFI_ERR_TRANSMIT);

			// Should get CLOSED as part of the returned data
			if(findStr(apduData.acReply,"CLOSED",6) == NULL)
				multosExitSW(WIFI_ERR_CLOSE);

			break;

		case WIFI_RESET_INS:
			// Disconnect Wifi
			// Clear last error
			break;

		case WIFI_MQTT_INS:
			if(!multosCheckCase(4))
				multosExitSW(ISO7816_SW_WRONG_PARAMS);
			miSerialIOFlushReceiveBuffer();

			if(P1 == MQTT_CONNECT)
			{
				// Build AT START command
				La = 0;
				acATCmd[0] = 0;
				strcat(acATCmd,"AT+CIPSTART=\"");
				strcat(acATCmd,apduData.mqttConnectData.acProtocol);
				strcat(acATCmd,"\",\"");
				strcat(acATCmd,apduData.mqttConnectData.acServer);
				strcat(acATCmd,"\",");
				ultoa(apduData.mqttConnectData.wPort,acATReply,10);
				strcat(acATCmd,acATReply);

				// Send AT command and check reply is OK
				wReplyLen = sendATCmd(acATCmd,acATReply,sizeof(acATReply),1000,TRUE);
				if(findStr(acATReply,"OK",2) == NULL)
				{
					La = 0;
					apduData.acReply[0] = 0;
					multosExitSW(WIFI_ERR_OPEN);
				}

				// Build AT SEND command
				acATCmd[0] = 0;
				strcat(acATCmd,"AT+CIPSEND=");
				bLen = strlen(apduData.mqttConnectData.acConnectionId);
				wReqLen = sizeof(mqttConnectHdr) + bLen;
				ultoa(wReqLen,acATReply,10);
				strcat(acATCmd,acATReply);

				// Send AT command and check reply is >
				wReplyLen = sendATCmd(acATCmd,acATReply,sizeof(acATReply),500,FALSE);
				if(wReplyLen < 2 || acATReply[wReplyLen-2] != '>')
				{
					La = 0;
					apduData.acReply[0] = 0;
					multosExitSW(WIFI_ERR_SEND);
				}

				// Set up the lengths in the header
				mqttConnectHdr[1] = wReqLen - 2;
				mqttConnectHdr[13] = bLen;

				// Transmit message
				miSerialIOTransmitBytes(mqttConnectHdr,sizeof(mqttConnectHdr));
				miSerialIOTransmitBytes((BYTE*)apduData.mqttConnectData.acConnectionId,bLen);
				wReplyLen = readReply(acATReply,sizeof(acATReply),1000,"+IPD,4");

				// Message doesn't return any data
				La = 0;
				apduData.acReply[0] = 0;

				// Check for valid CONNACK message
				if( wReplyLen < 4 || multosCompare((BYTE*)acATReply+wReplyLen-4,mqttConnAckMsg,sizeof(mqttConnAckMsg)) != MULTOS_BLOCK1_EQ_BLOCK2)
					multosExitSW(WIFI_ERR_TRANSMIT);
			}
			else if (P1 == MQTT_DISCONNECT)
			{
				// Build AT SEND command
				acATCmd[0] = 0;
				strcat(acATCmd,"AT+CIPSEND=2");

				// Send AT command and check reply is >
				wReplyLen = sendATCmd(acATCmd,acATReply,sizeof(acATReply),500,FALSE);
				if(wReplyLen < 2 || acATReply[wReplyLen-2] != '>')
				{
					La = 0;
					apduData.acReply[0] = 0;
					multosExitSW(WIFI_ERR_SEND);
				}

				// Transmit message
				miSerialIOTransmitBytes(mqttDisconnectMsg,sizeof(mqttDisconnectMsg));
				wReplyLen = readReply(acATReply,sizeof(acATReply),1000,"CLOSED");

				// Message doesn't return any data
				La = 0;
				apduData.acReply[0] = 0;

				// Should get CLOSED as part of the returned data
				if(findStr(acATReply,"CLOSED",6) == NULL)
					multosExitSW(WIFI_ERR_CLOSE);

			}
			else if (P1 == MQTT_PUBLISH)
			{
				// Build AT SEND command
				acATCmd[0] = 0;
				strcat(acATCmd,"AT+CIPSEND=");
				wTopicLen = strlen(apduData.mqttPublishData.acTopic);

				// Variable header + payload length
				// Topic Len (2) + Topic (var) + Payload (var)
				wReqLen = 2 + wTopicLen + apduData.mqttPublishData.wPayloadLen;
				if (wReqLen <= 127)
					// Fixed header fits in to two byte
					ultoa(wReqLen + 2,acATReply,10);
				else
					// Length field needs to be two bytes so total is three
					ultoa(wReqLen + 3,acATReply,10);
				strcat(acATCmd,acATReply);

				// Send AT command and check reply is >
				wReplyLen = sendATCmd(acATCmd,acATReply,sizeof(acATReply),500,FALSE);
				if(wReplyLen < 2 || acATReply[wReplyLen-2] != '>')
				{
					apduData.acReply[0] = 0;
					multosExitSW(WIFI_ERR_SEND);
				}

				// Transmit message
				miSerialIOTransmitBytes(mqttPublishHdr,1); // Message type
				if(wReqLen <= 127) // Remaining length
				{
					mqttPublishHdr[1] = wReqLen;
					miSerialIOTransmitBytes(mqttPublishHdr+1,1);
				}
				else
				{
					mqttPublishHdr[1] = wReqLen / 128;
					mqttPublishHdr[2] = wReqLen % 128;
					miSerialIOTransmitBytes(mqttPublishHdr+1,2);
				}
				miSerialIOTransmitBytes((BYTE*)&wTopicLen,2); // Topic length
				miSerialIOTransmitBytes((BYTE*)apduData.mqttPublishData.acTopic,wTopicLen); // Topic
				miSerialIOTransmitBytes(apduData.mqttPublishData.abPayload,apduData.mqttPublishData.wPayloadLen); // Payload
				wReplyLen = readReply(acATReply,sizeof(acATReply),1000, "SEND OK");

				// Message doesn't return any data
				La = 0;
				apduData.acReply[0] = 0;

				// Should get SEND OK as part of the returned data
				if(findStr(acATReply,"SEND OK",7) == NULL)
					multosExitSW(WIFI_ERR_TRANSMIT);
			}
			else
				multosExitSW(ISO7816_SW_INCORRECT_P1P2);
			break;

		default:
			multosExitSW(ISO7816_SW_INS_NOT_SUPPORTED);
			break;
	}
	multosExitSW(ISO7816_SW_NO_ERROR);
}

