/*
 * main.c
 *
 *  Created on: 19 Feb 2018
 *      Author: ctorr
 */

// *** RUN IN EMBEDDED MODE

#pragma attribute("aid", "F5000002")
#pragma attribute("access_list", "1003")  // Peripheral Access + Contact i/f + Strong Crypto
#pragma attribute("name", "Wifi Testbed")

#include <multos.h>
#include <mi.h>
#include <stdio.h>
#include <string.h>
#include "wifi.h"

// ****************** DEFINES ****************************
#define SERIAL_PORT_SPEED 57600
#define SERIAL_TX_GPIO_PIN 1
#define SERIAL_RX_GPIO_PIN 0
#define MAIN_SERIAL_PORT 0
#define SECONDARY_SERIAL_PORT 1

// **************** MEMORY SEGMENTS **********************
#pragma melpublic
union
{
	wifiConnectApdu_t connectData;
	wifiHttpGetReqApdu_t requestData;
	mqttConnectApdu_t mqttConnectData;
	mqttPublishApdu_t mqttPublishData;
	char acReply[WIFI_MAX_REPLY];
} apduData;


// ************* LOCAL FUNCTIONS ************************
void abendFnc(void)
{
	miSerialIOSetActivePort(MAIN_SERIAL_PORT);
	printf("Abend %d\n",miAbendReason());
	miPerformRestart();
}


// ***************** MAIN ******************************
// Send a one letter commands on the main serial port as follows:-
// i: Initialise wifi
// s: Test wifi connection
// q: MQTT test
void
main(void)
{
	char c;
	BYTE bEmbeddedEventId;
	WORD wStatus = 0x9000;

	bEmbeddedEventId = miGetEmbeddedEventId();

	// Initialise I/O on startup or reset
	if( bEmbeddedEventId == MI_EMBEDDED_EVENT_STARTUP )
	{
		// Register abend handler
		miSetAbendFunction((WORD)abendFnc);

		// Set up main serial port for receipt of commands and output of messages
		miSerialIOSetActivePort(MAIN_SERIAL_PORT);
		miSerialIOInitialise(MI_PACKAGE_PIN_27_GPIO_PIN_NUMBER,MI_PACKAGE_PIN_26_GPIO_PIN_NUMBER,9600,MI_SERIAL_IO_8_BITS_NO_PARITY_1_STOP_BIT, 100);
		miRegisterEmbeddedEvent(MI_EMBEDDED_EVENT_SERIAL_IO,4);
		printf("Hello\n");

		// Wait for ESP chip to get going
		miDelay(500);

		// Set up the other serial port to communicate with the ESP chip
		// Note that the buffer size has to be big enough to cope with the biggest message that can be returned. The AT software
		// on the ESP doesn't cater for sending the data back in chunks.
		miSerialIOSetActivePort(SECONDARY_SERIAL_PORT);
		miSerialIOInitialise(SERIAL_TX_GPIO_PIN,SERIAL_RX_GPIO_PIN,SERIAL_PORT_SPEED,MI_SERIAL_IO_8_BITS_NO_PARITY_1_STOP_BIT,1024);
	}

	// Handle command message
	else if (bEmbeddedEventId == MI_EMBEDDED_EVENT_SERIAL_IO)
	{
		// Receive command message
		miSerialIOSetActivePort(MAIN_SERIAL_PORT);
		miSerialIOGetReceivedBytes((BYTE*)&c,1);

		// Process message
		if(c == 'i')
		{
			// Ask the driver to initialise the Wifi module
			wStatus = wifiInit(WIFI_SERIAL_IF, SECONDARY_SERIAL_PORT);

			// If initialisation succeeded
			if(wStatus == 0x9000)
			{
				// Set up the WiFi connection information in public
				strcpy(apduData.connectData.acSSID,"YourAP");
				strcpy(apduData.connectData.acPWD,"YourPassword");

				// Call the library function to delegate to the driver
				wStatus = wifiBegin();
			}
			else
				wStatus = 0x9001;
		}
		else if (c == 's')
		{
			// Set up the HTTP GET request in public
			strcpy(apduData.requestData.acProtocol,"TCP");
			strcpy(apduData.requestData.acServer,"www.multos.com");
			apduData.requestData.wPort = 80;
			strcpy(apduData.requestData.acUri,"/testing.html");
			apduData.requestData.wMaxWait = 5000;

			// Call the library function to delegate to the driver
			wStatus = wifiHttpGet();

			// Print the reply from the driver contained in public
			miSerialIOSetActivePort(MAIN_SERIAL_PORT);
			printf("Received: %d\n", La);
			miSerialIOTransmitBytes((BYTE*)apduData.acReply,La);
		}
		else if (c == 'q')
		{
			// Set up the MQTT connection info in public
			strcpy(apduData.mqttConnectData.acProtocol,"TCP");
			strcpy(apduData.mqttConnectData.acServer,"127.0.0.1"); // Your MQTT IP address here
			apduData.mqttConnectData.wPort = 1883;
			apduData.mqttConnectData.wMaxWait = 5000;
			strcpy(apduData.mqttConnectData.acConnectionId,"test01");

			// Delegate to the driver to make the connection
			wStatus = mqttConnect();

			miSerialIOSetActivePort(MAIN_SERIAL_PORT);
			printf("Connect----------\n");
			printf("Status %04x\n", wStatus);

			// If connected to the MQTT server
			if(wStatus == 0x9000)
			{
				// Set up the topic and payload in public
				strcpy(apduData.mqttPublishData.acTopic,"test");
				strcpy((char *)apduData.mqttPublishData.abPayload,"Hello World");
				apduData.mqttPublishData.wPayloadLen = 11;

				// Delegate to the driver to publish the payload
				wStatus = mqttPublish();

				miSerialIOSetActivePort(MAIN_SERIAL_PORT);
				printf("Publish----------\n");
				printf("Status %04x\n", wStatus);
			}

			// Tidy up
			wStatus = mqttDisconnect();
			miSerialIOSetActivePort(MAIN_SERIAL_PORT);
			printf("Disconnect----------\n");
		}
		else
		{
			miSerialIOSetActivePort(MAIN_SERIAL_PORT);
			printf("Unknown command %c\n",c);
		}
		miSerialIOSetActivePort(MAIN_SERIAL_PORT);
		printf("\nStatus %04x\n", wStatus);
	}
}
